import functools
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from py65emu.cpu import CPU

InstructionConfigType = int | str | tuple[str, str | bool]

InstructionType = tuple[str, str, int, str, InstructionConfigType | None]


class UndefinedOperation(LookupError):
    pass


class Operation:
    cpu: "CPU"
    """CPU Object"""

    opcode: int
    """Opcode"""

    name: str
    """Operation mnemonic"""

    mode: str
    """Address mode for operation"""
    cycles: int
    """Number of cycles operation should take"""

    _type: str
    """Address mode type"""

    config: InstructionConfigType | None = None
    """Operation configuration"""

    def __init__(
        self,
        cpu: "CPU",
        opcode: int,
        name: str,
        mode: str,
        cycles: int,
        type: str,
        config: InstructionConfigType | None = None
    ):
        """
        Operation Object

        :param CPU cpu: CPU Object
        :param int opcode: Opcode
        :param str name: Operation mnemonic
        :param str mode: Address mode for operation
        :param int cycles: No. of cycles operation should take
        :param str type: Address mode type
        :param InstructionConfigType config: Operation configuration
        """
        self.cpu: "CPU" = cpu
        self.opcode = opcode
        self.name = name
        self.mode = mode
        self.cycles = cycles
        self.config = config
        self._type = type

    def target(self, config: InstructionConfigType) -> InstructionConfigType:
        """Wrapper method to return static configuration"""
        return config

    def execute(self) -> None:
        """Execute operation"""

        op_f = getattr(self.cpu, self.opname)
        if self.config:
            a_f = functools.partial(self.target, self.config)
        else:
            a_f = getattr(self.cpu, self.amode)

        op_f(a_f())

    def get_operands(self, addr: int | None = None) -> tuple[int, int, int]:
        """
        Read operands from memory

        :param addr: Address to read from, if None, read from CPU Register PC
        :type addr: int | None
        :rtype: tuple[int, int, int]
        :return: Tuple with address, hi-byte, lo-byte
        """

        if addr is None:
            addr = self.cpu.r.pc - 1

        hi = 0x00
        lo = 0x00
        if self.bytes == 2:
            lo = self.cpu.mmu.read(addr + 1)
        elif self.bytes == 3:
            lo = self.cpu.mmu.read(addr + 1)
            hi = self.cpu.mmu.read(addr + 2)

        return (addr, hi, lo)

    @property
    def bytes(self) -> int:
        """
        Number of bytes operation reads

        :meta private:
        """
        if self.mode in ['acc', 'imp']:
            return 1
        elif self.mode in ['im', 'z', 'zx', 'zy', 'ix', 'iy', 'rel']:
            return 2
        else:  # ['a', 'ax', 'ay', 'i']:
            return 3

    @property
    def amode(self) -> str:
        """
        Method name for Address Mode

        :meta private:
        """
        if self._type == "v":
            return self.mode

        return "{}_a".format(self.mode)

    @property
    def opname(self) -> str:
        """
        Method name for mnemonic

        :meta private:
        """

        if self.name in [
            "BPL", "BMI", "BVC", "BVS", "BCC", "BCS", "BNE", "BEQ"
        ]:
            return "B"
        elif self.name in ["CLC", "CLI", "CLV", "CLD"]:
            return "CL"
        elif self.name in ["SEC", "SEI", "SED"]:
            return "SE"
        elif self.name in ["PHA", "PLA", "PHP", "PLP"]:
            return "P"
        elif self.name in ["TAX", "TXA", "TAY", "TYA", "TXS", "TSX"]:
            return "T"
        elif self.name in ["SKB", "IGN"]:
            return "NOP"
        return self.name

    def __repr__(self) -> str:
        prefix = f"{self.opcode:0>2X}: {self.name} "

        match self.mode:
            case "acc":
                return prefix + "A "
            case "imp":
                return prefix
            case "im":
                return prefix + "#$BB "
            case "z":
                return prefix + "$LL "
            case "zx":
                return prefix + "$LL, X "
            case "zy":
                return prefix + "$LL, Y "
            case "a":
                return prefix + "$LLHH "
            case "ax":
                return prefix + "$LLHH, X "
            case "ay":
                return prefix + "$LLHH, Y "
            case "i":
                return prefix + "($LLHH) "
            case "ix":
                return prefix + "($LL, X) "
            case "iy":
                return prefix + "($LL), Y "
            case _:  # Actually: self.mode == 'rel'
                return prefix + "$BB [PC + $BB] "


class OpCodes:
    """
    OpCodes
    For each operation have the opcode as key, and 5-tuples containing the name
    of the operation, addressing mode, the base number of cycles it should
    take, whether it acts on values, "v" or on addresses, "a" and a target
    register, if valid.

    Addressing modes:

    =====  ====
    Func.  Description
    =====  ====
    acc    Accumulator (i.e CPU Register A)
    im     Immediate
    imp    Implied
    z      Zeropage
    zx     Zeropage, X-indexed
    zy     Zeropage, Y-indexed
    a      Absolute
    ax     Absolute, X-indexed
    ay     Absolute, Y-indexed
    i      Indirect
    ix     X-indexed, Indirect
    iy     Indirect, Y-indexed
    rel    Relative
    =====  ====
    """

    ops: list[Operation | None]

    def __init__(self, cpu: "CPU"):
        """
        Object to hold the instruction set

        :param CPU cpu: CPU Object
        """
        self.cpu = cpu
        self.ops = [None] * 0x100

        for opcode, config in self.instructions().items():
            if opcode not in self.ops:
                self.ops[opcode] = Operation(cpu, opcode, *config)

    def __getitem__(self, key: int) -> Operation:
        """
        Get a single operation

        :param int key: Opcode to get
        :rtype: Operation
        :return: Operation object for Opcode
        :raises: UndefinedOperation
        """
        op = next(
            (
                x for x in self.ops
                if isinstance(x, Operation) and x.opcode == key
            ),
            None
        )

        if isinstance(op, Operation):
            return op
        raise UndefinedOperation(
            'Operation {0:d} ({0:0>2X}) not instantiated'.format(key)
        )

    def instructions(self) -> dict[int, InstructionType]:
        """
        Returns 6502 instruction set (incl. illegal instructions)

        :rtype: dict[int, InstructionType]
        :returns: Full instruction set
        """
        return {
            # region Instructions 0x0"
            0x00: ("BRK", "imp", 7, "v", 1),
            0x01: ("ORA", "ix", 6, "v", None),
            0x02: ("KIL", "imp", 1, "v", 1),
            0x03: ("SLO", "ix", 8, "a", None),
            0x04: ("IGN", "z", 3, "a", None),
            0x05: ("ORA", "z", 3, "v", None),
            0x06: ("ASL", "z", 5, "a", None),
            0x07: ("SLO", "z", 5, "a", None),
            0x08: ("PHP", "imp", 3, "a", ("PH", "p")),
            0x09: ("ORA", "im", 2, "v", None),
            0x0A: ("ASL", "acc", 2, "a", "a"),
            0x0B: ("AAC", "im", 2, "v", None),  # ANC
            0x0C: ("IGN", "a", 4, "a", None),
            0x0D: ("ORA", "a", 4, "v", None),
            0x0E: ("ASL", "a", 6, "a", None),
            0x0F: ("SLO", "a", 6, "a", None),
            # endregion Instructions 0x0

            # region Instructions 0x1
            0x10: ("BPL", "rel", 2, "v", ("N", False)),
            0x11: ("ORA", "iy", 5, "v", None),
            0x12: ("KIL", "imp", 1, "v", 1),
            0x13: ("SLO", "iy", 8, "a", None),
            0x14: ("IGN", "zx", 4, "a", None),
            0x15: ("ORA", "zx", 4, "v", None),
            0x16: ("ASL", "zx", 6, "a", None),
            0x17: ("SLO", "zx", 6, "a", None),
            0x18: ("CLC", "im", 2, "v", "C"),
            0x19: ("ORA", "ay", 4, "v", None),
            0x1A: ("NOP", "imp", 2, "v", 1),
            0x1B: ("SLO", "ay", 7, "a", None),
            0x1C: ("IGN", "ax", 4, "a", None),
            0x1D: ("ORA", "ax", 4, "v", None),
            0x1E: ("ASL", "ax", 7, "a", None),
            0x1F: ("SLO", "ax", 7, "a", None),
            # endregion Instructions 0x1

            # region Instructions 0x2
            0x20: ("JSR", "a", 6, "a", None),
            0x21: ("AND", "ix", 6, "v", None),
            0x22: ("KIL", "imp", 1, "v", 1),
            0x23: ("RLA", "ix", 8, "a", None),
            0x24: ("BIT", "z", 3, "v", None),
            0x25: ("AND", "z", 3, "v", None),
            0x26: ("ROL", "z", 5, "a", None),
            0x27: ("RLA", "z", 5, "a", None),
            0x28: ("PLP", "imp", 4, "a", ("PL", "p")),
            0x29: ("AND", "im", 2, "v", None),
            0x2A: ("ROL", "acc", 2, "a", "a"),
            0x2B: ("AAC", "im", 2, "v", None),  # ANC
            0x2C: ("BIT", "a", 4, "v", None),
            0x2D: ("AND", "a", 4, "v", None),
            0x2E: ("ROL", "a", 6, "a", None),
            0x2F: ("RLA", "a", 6, "a", None),
            # endregion Instructions 0x2

            # region Instructions 0x3
            0x30: ("BMI", "rel", 2, "v", ("N", True)),
            0x31: ("AND", "iy", 5, "v", None),
            0x32: ("KIL", "imp", 1, "v", 1),
            0x33: ("RLA", "iy", 8, "a", None),
            0x34: ("IGN", "zx", 4, "a", None),
            0x35: ("AND", "zx", 4, "v", None),
            0x36: ("ROL", "zx", 6, "a", None),
            0x37: ("RLA", "zx", 6, "a", None),
            0x38: ("SEC", "imp", 2, "v", "C"),
            0x39: ("AND", "ay", 4, "v", None),
            0x3A: ("NOP", "imp", 2, "v", 1),
            0x3B: ("RLA", "ay", 7, "a", None),
            0x3C: ("IGN", "ax", 4, "a", None),
            0x3D: ("AND", "ax", 4, "v", None),
            0x3E: ("ROL", "ax", 7, "a", None),
            0x3F: ("RLA", "ax", 7, "a", None),
            # endregion Instructions 0x3

            # region Instructions 0x4
            0x40: ("RTI", "imp", 6, "a", 1),
            0x41: ("EOR", "ix", 6, "v", None),
            0x42: ("KIL", "imp", 1, "v", 1),
            0x43: ("SRE", "ix", 8, "a", None),
            0x44: ("IGN", "z", 3, "a", None),
            0x45: ("EOR", "z", 3, "v", None),
            0x46: ("LSR", "z", 5, "a", None),
            0x47: ("SRE", "z", 5, "a", None),
            0x48: ("PHA", "imp", 3, "a", ("PH", "a")),
            0x49: ("EOR", "im", 2, "v", None),
            0x4A: ("LSR", "acc", 2, "a", "a"),
            0x4B: ("ASR", "im", 2, "v", None),  # ALR
            0x4C: ("JMP", "a", 3, "a", None),
            0x4D: ("EOR", "a", 4, "v", None),
            0x4E: ("LSR", "a", 6, "a", None),
            0x4F: ("SRE", "a", 6, "a", None),
            # endregion Instructions 0x4

            # region Instructions 0x5
            0x50: ("BVC", "rel", 2, "v", ("V", False)),
            0x51: ("EOR", "iy", 5, "v", None),
            0x52: ("KIL", "imp", 1, "v", 1),
            0x53: ("SRE", "iy", 8, "a", None),
            0x54: ("IGN", "zx", 4, "a", None),
            0x55: ("EOR", "zx", 4, "v", None),
            0x56: ("LSR", "zx", 6, "a", None),
            0x57: ("SRE", "zx", 6, "a", None),
            0x58: ("CLI", "im", 2, "v", "I"),
            0x59: ("EOR", "ay", 4, "v", None),
            0x5A: ("NOP", "imp", 2, "v", 1),
            0x5B: ("SRE", "ay", 7, "a", None),
            0x5C: ("IGN", "ax", 4, "a", None),
            0x5D: ("EOR", "ax", 4, "v", None),
            0x5E: ("LSR", "ax", 7, "a", None),
            0x5F: ("SRE", "ax", 7, "a", None),
            # endregion Instructions 0x5

            # region Instructions 0x6
            0x60: ("RTS", "imp", 6, "a", 1),
            0x61: ("ADC", "ix", 6, "v", None),
            0x62: ("KIL", "imp", 1, "v", 1),
            0x63: ("RRA", "ix", 8, "a", None),
            0x64: ("IGN", "z", 3, "a", None),
            0x65: ("ADC", "z", 3, "v", None),
            0x66: ("ROR", "z", 5, "a", None),
            0x67: ("RRA", "z", 5, "a", None),
            0x68: ("PLA", "imp", 4, "a", ("PL", "a")),
            0x69: ("ADC", "im", 2, "v", None),
            0x6A: ("ROR", "acc", 2, "a", "a"),
            0x6B: ("ARR", "im", 2, "v", None),
            0x6C: ("JMP", "i", 5, "a", None),
            0x6D: ("ADC", "a", 4, "v", None),
            0x6E: ("ROR", "a", 6, "a", None),
            0x6F: ("RRA", "a", 6, "a", None),
            # endregion Instructions 0x6

            # region Instructions 0x7
            0x70: ("BVS", "rel", 2, "v", ("V", True)),
            0x71: ("ADC", "iy", 5, "v", None),
            0x72: ("KIL", "imp", 1, "v", 1),
            0x73: ("RRA", "iy", 8, "a", None),
            0x74: ("IGN", "zx", 4, "a", None),
            0x75: ("ADC", "zx", 4, "v", None),
            0x76: ("ROR", "zx", 6, "a", None),
            0x77: ("RRA", "zx", 6, "a", None),
            0x78: ("SEI", "imp", 2, "v", "I"),
            0x79: ("ADC", "ay", 4, "v", None),
            0x7A: ("NOP", "imp", 2, "v", 1),
            0x7B: ("RRA", "ay", 7, "a", None),
            0x7C: ("IGN", "ax", 4, "a", None),
            0x7D: ("ADC", "ax", 4, "v", None),
            0x7E: ("ROR", "ax", 7, "a", None),
            0x7F: ("RRA", "ax", 7, "a", None),
            # endregion Instructions 0x7

            # region Instructions 0x8
            0x80: ("SKB", "im", 2, "v", None),
            0x81: ("STA", "ix", 6, "a", None),
            0x82: ("SKB", "im", 2, "v", None),
            0x83: ("AAX", "ix", 6, "a", None),  # SAX
            0x84: ("STY", "z", 3, "a", None),
            0x85: ("STA", "z", 3, "a", None),
            0x86: ("STX", "z", 3, "a", None),
            0x87: ("AAX", "z", 3, "a", None),  # SAX
            0x88: ("DEY", "imp", 2, "a", 1),
            0x89: ("SKB", "im", 2, "v", None),
            0x8A: ("TXA", "imp", 2, "a", ('x', 'a')),
            0x8B: ("XAA", "im", 2, "v", None),
            0x8C: ("STY", "a", 4, "a", None),
            0x8D: ("STA", "a", 4, "a", None),
            0x8E: ("STX", "a", 4, "a", None),
            0x8F: ("AAX", "a", 4, "a", None),  # SAX
            # endregion Instructions 0x8

            # region Instructions 0x9
            0x90: ("BCC", "rel", 2, "v", ("C", False)),
            0x91: ("STA", "iy", 6, "a", None),
            0x92: ("KIL", "imp", 1, "v", 1),
            0x93: ("AXA", "iy", 6, "a", None),  # SHA
            0x94: ("STY", "zx", 4, "a", None),
            0x95: ("STA", "zx", 4, "a", None),
            0x96: ("STX", "zy", 4, "a", None),
            0x97: ("AAX", "zy", 4, "a", None),  # SAX
            0x98: ("TYA", "imp", 2, "a", ('y', 'a')),
            0x99: ("STA", "ay", 5, "a", None),
            0x9A: ("TXS", "imp", 2, "a", ('x', 's')),
            0x9B: ("XAS", "ay", 5, "a", None),  # SHS, TAS
            0x9C: ("SYA", "ax", 5, "a", None),
            0x9D: ("STA", "ax", 5, "a", None),
            0x9E: ("SXA", "ay", 5, "a", None),
            0x9F: ("AXA", "ay", 5, "a", None),  # SHA
            # endregion Instructions 0x9

            # region Instructions 0xA
            0xA0: ("LDY", "im", 2, "v", None),
            0xA1: ("LDA", "ix", 6, "v", None),
            0xA2: ("LDX", "im", 2, "v", None),
            0xA3: ("LAX", "ix", 6, "v", None),
            0xA4: ("LDY", "z", 3, "v", None),
            0xA5: ("LDA", "z", 3, "v", None),
            0xA6: ("LDX", "z", 3, "v", None),
            0xA7: ("LAX", "z", 3, "v", None),
            0xA8: ("TAY", "imp", 2, "a", ('a', 'y')),
            0xA9: ("LDA", "im", 2, "v", None),
            0xAA: ("TAX", "imp", 2, "a", ('a', 'x')),
            0xAB: ("ATX", "im", 2, "v", None),  #
            0xAC: ("LDY", "a", 4, "v", None),
            0xAD: ("LDA", "a", 4, "v", None),
            0xAE: ("LDX", "a", 4, "v", None),
            0xAF: ("LAX", "a", 4, "v", None),
            # endregion Instructions 0xA

            # region Instructions 0xB
            0xB0: ("BCS", "rel", 2, "v", ("C", True)),
            0xB1: ("LDA", "iy", 5, "v", None),
            0xB2: ("KIL", "imp", 1, "v", 1),
            0xB3: ("LAX", "iy", 5, "v", None),
            0xB4: ("LDY", "zx", 4, "v", None),
            0xB5: ("LDA", "zx", 4, "v", None),
            0xB6: ("LDX", "zy", 4, "v", None),
            0xB7: ("LAX", "zy", 4, "v", None),
            0xB8: ("CLV", "imp", 2, "v", "V"),
            0xB9: ("LDA", "ay", 4, "v", None),
            0xBA: ("TSX", "imp", 2, "a", ('s', 'x')),
            0xBB: ("LAR", "ay", 4, "v", None),
            0xBC: ("LDY", "ax", 4, "v", None),
            0xBD: ("LDA", "ax", 4, "v", None),
            0xBE: ("LDX", "ay", 4, "v", None),
            0xBF: ("LAX", "ay", 4, "v", None),
            # endregion Instructions 0xB

            # region Instructions 0xC
            0xC0: ("CPY", "im", 2, "v", None),
            0xC1: ("CMP", "ix", 6, "v", None),
            0xC2: ("SKB", "im", 2, "v", None),
            0xC3: ("DCP", "ix", 8, "a", None),
            0xC4: ("CPY", "z", 3, "v", None),
            0xC5: ("CMP", "z", 3, "v", None),
            0xC6: ("DEC", "z", 5, "a", None),
            0xC7: ("DCP", "z", 5, "a", None),
            0xC8: ("INY", "imp", 2, "a", 1),
            0xC9: ("CMP", "im", 2, "v", None),
            0xCA: ("DEX", "imp", 2, "a", 1),
            0xCB: ("AXS", "im", 2, "v", None),  # SBX
            0xCC: ("CPY", "a", 4, "v", None),
            0xCD: ("CMP", "a", 4, "v", None),
            0xCE: ("DEC", "a", 6, "a", None),
            0xCF: ("DCP", "a", 6, "a", None),
            # endregion Instructions 0xC

            # region Instructions 0xD
            0xD0: ("BNE", "rel", 2, "v", ("Z", False)),
            0xD1: ("CMP", "iy", 5, "v", None),
            0xD2: ("KIL", "imp", 1, "v", 1),
            0xD3: ("DCP", "iy", 8, "a", None),
            0xD4: ("IGN", "zx", 4, "a", None),
            0xD5: ("CMP", "zx", 4, "v", None),
            0xD6: ("DEC", "zx", 6, "a", None),
            0xD7: ("DCP", "zx", 6, "a", None),
            0xD8: ("CLD", "imp", 2, "v", "D"),
            0xD9: ("CMP", "ay", 4, "v", None),
            0xDA: ("NOP", "imp", 2, "v", 1),
            0xDB: ("DCP", "ay", 7, "a", None),
            0xDC: ("IGN", "ax", 4, "a", None),
            0xDD: ("CMP", "ax", 4, "v", None),
            0xDE: ("DEC", "ax", 7, "a", None),
            0xDF: ("DCP", "ax", 7, "a", None),
            # endregion Instructions 0xD

            # region Instructions 0xE
            0xE0: ("CPX", "im", 2, "v", None),
            0xE1: ("SBC", "ix", 6, "v", None),
            0xE2: ("SKB", "im", 2, "v", None),
            0xE3: ("ISC", "ix", 8, "a", None),
            0xE4: ("CPX", "z", 3, "v", None),
            0xE5: ("SBC", "z", 3, "v", None),
            0xE6: ("INC", "z", 5, "a", None),
            0xE7: ("ISC", "z", 5, "a", None),
            0xE8: ("INX", "imp", 2, "a", 1),
            0xE9: ("SBC", "im", 2, "v", None),
            0xEA: ("NOP", "imp", 2, "v", 1),
            0xEB: ("SBC", "im", 2, "v", None),  # USBC
            0xEC: ("CPX", "a", 4, "v", None),
            0xED: ("SBC", "a", 4, "v", None),
            0xEE: ("INC", "a", 6, "a", None),
            0xEF: ("ISC", "a", 6, "a", None),
            # endregion Instructions 0xE

            # region Instructions 0xF
            0xF0: ("BEQ", "rel", 2, "v", ("Z", True)),
            0xF1: ("SBC", "iy", 5, "v", None),
            0xF2: ("KIL", "imp", 1, "v", 1),
            0xF3: ("ISC", "iy", 8, "a", None),
            0xF4: ("IGN", "zx", 4, "a", None),
            0xF5: ("SBC", "zx", 4, "v", None),
            0xF6: ("INC", "zx", 6, "a", None),
            0xF7: ("ISC", "zx", 6, "a", None),
            0xF8: ("SED", "imp", 2, "v", "D"),
            0xF9: ("SBC", "ay", 4, "v", None),
            0xFA: ("NOP", "imp", 2, "v", 1),
            0xFB: ("ISC", "ay", 7, "a", None),
            0xFC: ("IGN", "ax", 4, "a", None),
            0xFD: ("SBC", "ax", 4, "v", None),
            0xFE: ("INC", "ax", 7, "a", None),
            0xFF: ("ISC", "ax", 7, "a", None),
            # endregion Instructions 0xF
        }



from typing import TYPE_CHECKING
import math

if TYPE_CHECKING:
    from py65emu.operation import Operation
    from py65emu.cpu import CPU


class Disassembly:
    """
    Disassembly class, to help disassemble a program in memory
    """

    def __init__(
        self,
        op: "Operation",
        pc: int,
        hi: int = 0x00,
        lo: int = 0x00
    ):
        """
        Init Disassembly Method

        :param Operation op: Operation to disassemble
        :param int pc: Current programcounter
        :param int hi: Hi-byte of memory. (Default: `0x00`)
        :param int lo: Lo-byte of memory. (Default: `0x00`)
        """
        self.op = op
        self.pc = pc
        self.hi = hi
        self.lo = lo

    def as_word(self) -> int:
        """Returns hi & lo byte as word (16 bit)"""
        return (self.hi << 8) + self.lo

    @property
    def memory(self) -> str:
        """
        Returns mnemonic string with current memory addresses
        """
        prefix = f"{self.op.name} "
        match self.op.mode:
            case "acc":
                return prefix + "A "
            case "imp":
                return prefix
            case "im":
                return prefix + f"#${self.lo:0>2x} "
            case "z":
                return prefix + f"${self.lo:0>2x} "
            case "zx":
                addr = (self.lo + self.op.cpu.r.x) & 0xFFFF
                return prefix + f"${self.lo:0>2x}, X [${addr:0>4x}]"
            case "zy":
                addr = (self.lo + self.op.cpu.r.y) & 0xFFFF
                return prefix + f"${self.lo:0>2x}, Y [${addr:0>4x}]"
            case "a":
                return prefix + f"${self.hi:0>2x}{self.lo:0>2x} "
            case "ax":
                addr = (self.as_word() + self.op.cpu.r.x) & 0xFFFF
                return prefix + (
                    f"${self.hi:0>2x}{self.lo:0>2x}, "
                    f"X [${addr:0>4x}]"
                )
            case "ay":
                addr = (self.as_word() + self.op.cpu.r.y) & 0xFFFF
                return prefix + (
                    f"${self.hi:0>2x}{self.lo:0>2x}, "
                    f"Y [${addr:0>4x}]"
                )
            case "i":
                return prefix + f"(${self.hi:0>2x}{self.lo:0>2x}) "
            case "ix":
                loc_addr = (self.lo + self.op.cpu.r.x) & 0xFFFF
                addr = self.op.cpu.mmu.readWord(loc_addr) & 0xFFFF
                return prefix + f"(${self.lo:0>2x}, X) [${addr:0>4x}]"
            case "iy":
                loc_addr = self.op.cpu.mmu.readWord(self.lo) & 0xFFFF
                addr = loc_addr + self.op.cpu.r.y
                return prefix + f"(${self.lo:0>2x}), Y [${addr:0>4x}]"
            case _:  # Actually: self.op.mode == 'rel'
                addr = (
                    (self.pc + 1) + self.op.cpu.fromTwosCom(self.lo)
                ) & 0xFFFF
                return prefix + f"${self.lo:0>2x} [${addr:0>4x}] "

    def __repr__(self) -> str:
        """Returns everything nicely formated"""
        return f"${self.pc:0>4x} {self.op.opcode:0>2x} {self.lo:0>2x} "\
               f"{self.hi:0>2x} {self.op.opname: >3s}: {self.memory}"


class Debug:
    """
    Debug class, do help debug a program or the module itself.
    """

    def __init__(self, cpu: "CPU"):
        """
        :param CPU cpu: The CPU object to debug
        """
        self.cpu = cpu

    def _get_assembly(self, addr: int) -> tuple[int, "Disassembly"]:
        """
        This is the disassembly function.
        Its workings are not required for emulation.

        It is merely a convenience function to turn the binary instruction
        code into human readable form. Its included as part of the emulator
        because it can take advantage of many of the CPUs internal operations
        to do

        :mode private:
        :param int addr: address to disassemble
        :rtype: tuple[int, Disassembly]
        :return: Tuple with the current address and corresponding
                 Disassembly object
        """
        addr_org = addr
        opcode = self.cpu.mmu.read(addr)
        addr += 1
        hi = 0x00
        lo = 0x00

        op = self.cpu.opcodes[opcode]
        if op.bytes == 2:
            lo = self.cpu.mmu.read(addr)
            addr += 1
        elif op.bytes == 3:
            lo = self.cpu.mmu.read(addr)
            addr += 1
            hi = self.cpu.mmu.read(addr)
            addr += 1

        assembly = Disassembly(
            op=op,
            pc=addr_org,
            hi=hi,
            lo=lo
        )

        return addr, assembly

    def _get_memory(self, start: int, stop: int) -> list[tuple[int, ...]]:
        """
        Dumps a portion of the memory.
        This reads over blocks.

        :mode private:
        :param int start: Start offset
        :param int stop: Stop offset
        :rtype: list[tuple[int, ...]]
        :return: A list with a tuple of the memory data from the selected
                 memory span
        """
        start_offset = start & 0xFFF0
        stop_offset = stop | 0x000F

        offset_length = math.ceil((stop_offset - start_offset) / 16)

        memory = []

        for multiplier in range(offset_length):
            offset = start_offset + (multiplier * 0x0010)

            value = [offset,]
            for addr in range(0x10):
                value += self.cpu.mmu.read(offset + addr),

            memory.append(tuple(value))
        return memory

    def _disassemble_params(
        self, start: int | None = None, stop: int | None = None
    ) -> tuple[int, int, int]:
        if start is None:
            start = self.cpu.r.pc - 10
        if start < 0:
            start = 0

        addr = start & 0xFFFF

        if stop is None:
            stop = self.cpu.r.pc
        if stop < start:
            stop = start

        return start, stop, addr

    def d(self, addr: int) -> None:
        """
        Shorthand method for print out dump of memory and disassembly for
        selected address

        :param addr: Address to dump
        :type addr: int
        """
        self.memdump(addr)
        self.disassemble(addr)

    @staticmethod
    def crash_dump(cpu: "CPU") -> None:
        """
        Static method to dump out backtrace of the program that was running
        """
        entry = cpu.mmu.read(0xFFFD) + (cpu.mmu.read(0xFFFC) << 8)

        pc = cpu.r.pc - 20
        if pc < entry:
            pc = entry

        d = Debug(cpu)
        d.disassemble(pc, cpu.r.pc)

    """Disassembly methods."""
    def disassemble(
        self, start: int | None = None, stop: int | None = None
    ) -> None:
        """
        :param start: The starting address. (Default: PC - 20)
        :param stop: The stopping address. If stop is lower than start then
                     parameter will be used as "length" (ie. incr. by start)
                     instead. (Default: PC)
        :type start: int | None
        :type stop: int | None
        """
        start, stop, addr = self._disassemble_params(start, stop)

        print(
            "DISASSEMBLE: ${:0>4x} - ${:0>4x}\n"
            "OP LO HI OPS DISASSEMBLY"
            .format(start, stop)
        )

        while addr <= (stop & 0xFFFF):
            addr, obj = self._get_assembly(addr)
            print(f"{obj!r}")

    def disassemble_list(
        self, start: int | None = None, stop: int | None = None
    ) -> list[Disassembly]:
        """
        See :py:meth:`py65emu.debug.Debug.disassemble` for a more descriptive
        text

        .. seealso::
           :py:meth:`py65emu.debug.Debug.disassemble`

        :param start: The starting address. (Default: PC - 20)
        :param stop: The stopping address. If stop is lower than start then
                     parameter will be used as "length" (ie. incr. by start)
                     instead. (Default: PC)
        :type start: int | None
        :type stop: int | None
        :rtype: list[Disassembly]
        :return: A list of Disassembly objects
        """
        start, stop, addr = self._disassemble_params(start, stop)

        lines: list[Disassembly] = []
        while addr <= (stop & 0xFFFF):
            cur = addr
            addr, assembly = self._get_assembly(addr)
            lines.insert(cur, assembly)
        return lines

    """Memory methods."""
    def memdump(self, start: int, stop: int | None = None) -> None:
        """
        Prints the memory data between `start` and `stop` parameters.
        If `stop` parameter is left out, it will dump the 16 byte segment where
        the `start` address exists.
        (`start = start & 0xFFF0` and `stop = start | 0x000F`)

        This method reads over blocks.

        :param int start: Start offset
        :param int | None stop: Stop offset. (Default: None)
        """
        offset = None
        if stop is None:
            offset = ((start | 0x000F) - ((start & 0xFFF0)) + 1)
            start = start & 0xFFF0
            stop = (start | 0x000F)

        print(
            f"MEMORY DUMP FOR: ${start:0>4x} - ${stop:0>4x}\n"
            "ADDR 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F"
        )

        memory = self._get_memory(start, stop)
        output = "{:0>4x}"
        output += " {:0>2x}" * 16

        for row in memory:
            print(output.format(*row))

        if offset is not None:
            print("".ljust((offset * 3) + 1, " ") + " ^^")

    def stackdump(self, pointer: int | None = None) -> None:
        """
        Convience method to dump the stack around the current stackpointer
        (Register S)

        :param pointer: Stack Pointer Location
        :type pointer: int | None
        """
        if not pointer:
            pointer = self.cpu.r.s

        self.memdump((self.cpu.stack_page * 0x100) + pointer)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_timing
----------------------------------

Timing tests for `py65emu` module.
"""


import os
import unittest
import csv

from py65emu.cpu import CPU
from py65emu.mmu import MMU


class Timing(unittest.TestCase):
    def setUp(self):
        pass

    def load_csv_data(self) -> dict[int, dict[str, object]]:
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            "6502_cycle_test.csv"
        )

        cycle_test_data_results = {}
        with open(path) as fp:
            test_data = csv.reader(fp, delimiter=",", quotechar='"')
            line_number = 0
            for row in test_data:
                if int(row[0]) % 2 != 0:
                    line_number += 1
                    continue

                if row[8].strip() == "":
                    continue

                cycle_test_data_results[line_number] = {
                    "PC": int(row[1], 16),
                    "A": int(row[2], 16),
                    "X": int(row[3], 16),
                    "Y": int(row[4], 16),
                    "P": int(row[5], 16),
                    "SP": int(row[6], 16),
                    "CC": int(row[7]),
                    "CMD": row[8],
                }

        return cycle_test_data_results

    def test_cycle_test(self):
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            "6502_cycle_test.bin"
        )

        mmu = MMU()
        with open(path, "rb") as f:
            mmu.addBlock(0x0000, 0x10000, False, f)

        c = CPU(mmu, 0x00)
        c.r.s = 0xFD

        total_no_of_cycles = 0
        loop = 1

        cycle_test_data_results = self.load_csv_data()

        while 1:
            # c.step()
            # total_no_of_cycles = total_no_of_cycles + c.cc

            if loop not in cycle_test_data_results:
                loop += 1
                continue

            self.assertEqual(
                c.r.pc,
                cycle_test_data_results[loop]["PC"],
                "Step: {:d} PC: {:0>4x} - {} - {}".format(
                    loop,
                    c.r.pc,
                    cycle_test_data_results[loop]["CMD"],
                    c.r
                )
            )
            self.assertEqual(
                total_no_of_cycles,
                cycle_test_data_results[loop]["CC"],
                "Step: {:d} Cycle: {:d} - {} - {}".format(
                    loop,
                    total_no_of_cycles,
                    cycle_test_data_results[loop]["CMD"],
                    c.r,
                )
            )
            c.step()
            total_no_of_cycles += c.cc

            loop += 1

            if c.r.pc == 0x1266:
                break

            self.assertLess(
                total_no_of_cycles, 1147, "Maximum number of cycles exceeded"
            )
            self.assertLess(loop, 2295, "Maximum number of loops exceeded")

    def tearDown(self):
        pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_suites
----------------------------------

Tests for `py65emu` module.
"""

import unittest
import io
import unittest.mock

from typing import Sequence
from py65emu.cpu import CPU
from py65emu.debug import Disassembly, Debug
from py65emu.mmu import MMU


class BaseDebug(unittest.TestCase):
    subtests = [
        # OPC   LO    HI
        (0xEA, 0x00, 0x00, 'NOP'),
        (0x0A, 0x00, 0x00, 'ASL A'),
        (0xA9, 0x38, 0x00, 'LDA #$38'),
        (0xAD, 0x00, 0x00, 'LDA $0000'),
        (0xAD, 0x01, 0x02, 'LDA $0201'),
        (0xA5, 0x34, 0x12, 'LDA $34'),
        (0xB5, 0x34, 0x12, 'LDA $34, X [$0035]'),
        (0xB6, 0x34, 0x12, 'LDX $34, Y [$0036]'),
        (0xAD, 0x34, 0x12, 'LDA $1234'),
        (0xBD, 0x34, 0x12, 'LDA $1234, X [$1235]'),
        (0xB9, 0x00, 0x12, 'LDA $1200, Y [$1202]'),
        (0xA1, 0x01, 0x00, 'LDA ($01, X) [$0000]'),
        (0xB1, 0x00, 0x00, 'LDA ($00), Y [$00b3]'),
        (0x6C, 0x01, 0x00, 'JMP ($0001)'),
        (0x6C, 0x01, 0x02, 'JMP ($0201)'),
        (0x6C, 0x00, 0x00, 'JMP ($0000)'),
        (0x90, 0x0C, 0x00, 'BCC $0c [$000d]'),
        (0xB0, 0x90, 0x00, 'BCS $90 [$ff91]'),
    ]

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _cpu(self, program=None, pc=0x0000, **kwargs) -> CPU:

        self.c = CPU(
            mmu=MMU([(0x0, 0x10000, False, program, pc)]),
            pc=pc,
            **kwargs
        )
        self.c.r.a = 0xAA
        self.c.r.x = 0x01
        self.c.r.y = 0x02

        return self.c

    def data_to_str(self, data: Sequence, pc: int | None = None) -> str:
        o = self.c.opcodes[data[0]]
        op = {
            "pc":     pc if pc is not None else self.c.r.pc,
            "opc":    data[0],
            "memory": data[3],
            "lo":     0x00,
            "hi":     0x00,
        }
        if o.bytes == 2:
            op['lo'] = data[1]
        elif o.bytes == 3:
            op['lo'] = data[1]
            op['hi'] = data[2]

        val = f'${op["pc"]:0>4x} {op["opc"]:0>2x} {op["lo"]:0>2x} '\
              f'{op["hi"]:0>2x} {o.opname: >3s}: {op["memory"]}'
        return val.strip()

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def assert_stdout(self, n, expected_output, mock_stdout):
        call = getattr(n[0], n[1])
        call()
        self.assertIn(mock_stdout.getvalue(), expected_output)

    def test_debug_active(self):
        # OPC   LO    HI
        data = (0xEA, 'NOP')
        c = self._cpu(program=[data[0],], debug=True)
        expected_pc = (c.r.pc + 1)
        expected = (
            '${r.pc:0>4x} {data[0]:0>2x} 00 00 {data[1]:s}: '
            '{data[1]: <24s} A: {r.a:0>2X} X: {r.x:0>2X} Y: {r.y:0>2X} '
            'S: {r.s:0>2X} PC: {pc1:0>4X} P: {r.p:0>2X}'
        )
        v = expected.format(
                data=data,
                pc1=expected_pc,
                r=c.r,
            )

        with unittest.mock.patch(
            'sys.stdout', new_callable=io.StringIO
        ) as mock_stdout:
            c.step()
            self.assertIn(
                v,
                mock_stdout.getvalue().strip(),
            )


class TestDisassembly(BaseDebug):
    def _dasm(self, **kwargs) -> Disassembly:
        self._cpu(**kwargs)
        opc = self.c.nextByte()
        op = self.c.opcodes[opc]

        dasm = Disassembly(op, *op.get_operands())
        return dasm

    def test_initialized_correct(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dasm = self._dasm(program=[data[0], data[1], data[2]])

                self.assertIsInstance(dasm, Disassembly)

                self.assertEqual(dasm.op.opcode, data[0], f'0x{data[0]:0>2X}')
                if dasm.op.bytes == 2:
                    self.assertEqual(dasm.lo, data[1], f'0x{data[0]:0>2X}')
                elif dasm.op.bytes == 3:
                    self.assertEqual(dasm.lo, data[1], f'0x{data[0]:0>2X}')
                    self.assertEqual(dasm.hi, data[2], f'0x{data[0]:0>2X}')

    def test_as_word(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dasm = self._dasm(program=[data[0], data[1], data[2]])
                dasm.lo = data[1]
                dasm.hi = data[2]

                word = (data[2] << 8) + data[1]

                self.assertEqual(dasm.op.opcode, data[0], f'0x{data[0]:0>2X}')
                self.assertEqual(dasm.as_word(), word, f'0x{data[0]:0>2X}')

    def test_memory_string(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dasm = self._dasm(program=[data[0], data[1], data[2]])

                self.assertEqual(dasm.op.opcode, data[0], f'0x{data[0]:0>2X}')
                self.assertEqual(
                    dasm.memory.strip(),
                    data[3],
                    f'0x{data[0]:0>2X}'
                )

    def test_repr_string(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dasm = self._dasm(program=[data[0], data[1], data[2]])

                self.assertEqual(dasm.op.opcode, data[0], f'0x{data[0]:0>2X}')
                self.assertEqual(
                    repr(dasm).strip(),
                    self.data_to_str(data, 0x0000),
                    f'0x{data[0]:0>2X}'
                )


class TestDebug(BaseDebug):
    def _debug(self, **kwargs) -> Debug:
        self._cpu(**kwargs)
        debug = Debug(self.c)

        return debug

    def memdump_str(
        self,
        addr: int,
        data: tuple[int, int, int, str] | None = None,
        suffix: bool = False
    ) -> str:
        memory_prefix = (
            "MEMORY DUMP FOR: ${start:0>4x} - ${stop:0>4x}\n"
            "ADDR 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F\n"
        )
        memory_suffix = " {offset: >{width}} ^^"

        res = [0x0] * 17
        res[0] = addr
        if data is not None and len(data) > 0:
            for k in range(3):
                val = data[k]
                if type(val) is int:
                    res[(k + 1)] = val

        row = "{:0>4x}"
        row += " {:0>2x}" * 16

        t = memory_prefix + row.format(*res)
        if suffix is True:
            t += "\n" + memory_suffix
        return t

    def test_initialized_correct(self):
        for data in self.subtests:
            with self.subTest(data=data):
                debug = self._debug(program=[data[0], data[1], data[2]])

                self.assertIsInstance(debug, Debug)
                self.assertIsInstance(debug.cpu, CPU)

    def test_get_assembly(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dbg = self._debug(program=[data[0], data[1], data[2]])
                o = self.c.opcodes[data[0]]

                bytes, dasm = dbg._get_assembly(self.c.r.pc)

                self.assertEqual(
                    o.opcode,
                    data[0],
                    f'0x{data[0]:0>2X} - OPC'
                )
                self.assertEqual(
                    o.bytes,
                    bytes,
                    f'0x{data[0]:0>2X} - Bytes'
                )
                self.assertEqual(
                    repr(dasm).strip(),
                    self.data_to_str(data),
                    f'0x{data[0]:0>2X} - Repr.'
                )

    def test_get_memory(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dbg = self._debug(program=[data[0], data[1], data[2]])

                memory = dbg._get_memory(0x0000, 0x0003)
                mem = memory.pop()

                self.assertEqual(mem[0], 0x00)  # Memory Offset
                self.assertEqual(mem[1], data[0])
                self.assertEqual(mem[2], data[1])
                self.assertEqual(mem[3], data[2])

    def test_shorthand_d(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dbg = self._debug(program=[data[0], data[1], data[2]])
                self.c.r.pc = 0x0000

                with unittest.mock.patch(
                    'sys.stdout', new_callable=io.StringIO
                ) as mock_stdout:
                    dbg.d(0x0000)
                    self.assertIsNotNone(mock_stdout.getvalue())

    def test_crash_dump(self):
        for data in self.subtests:
            with self.subTest(data=data):
                cpu = self._cpu(program=[data[0], data[1], data[2]])
                self.c.r.pc = 0x0000

                v = 'DISASSEMBLE: $0000 - $0000\n'\
                    'OP LO HI OPS DISASSEMBLY\n'\
                    + self.data_to_str(data)

                with unittest.mock.patch(
                    'sys.stdout', new_callable=io.StringIO
                ) as mock_stdout:
                    Debug.crash_dump(cpu)
                    self.assertIn(mock_stdout.getvalue().strip(), v)

    def test_disassemble(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dbg = self._debug(program=[data[0], data[1], data[2]])

                start = 0
                stop = 0
                v = f'DISASSEMBLE: ${start:0>4x} - ${stop:0>4x}\n'\
                    'OP LO HI OPS DISASSEMBLY\n' + self.data_to_str(data)

                with unittest.mock.patch(
                    'sys.stdout', new_callable=io.StringIO
                ) as mock_stdout:
                    dbg.disassemble()
                    self.assertIn(mock_stdout.getvalue().strip(), v)

    def test_disassemble_with_negative_stop(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dbg = self._debug(program=[data[0], data[1], data[2]])

                start = 0
                stop = 0
                v = f'DISASSEMBLE: ${start:0>4x} - ${stop:0>4x}\n'\
                    'OP LO HI OPS DISASSEMBLY\n' + self.data_to_str(data)
                stop = -1

                with unittest.mock.patch(
                    'sys.stdout', new_callable=io.StringIO
                ) as mock_stdout:
                    dbg.disassemble(start, stop)
                    self.assertIn(mock_stdout.getvalue().strip(), v)

    def test_disassemble_list(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dbg = self._debug(program=[data[0], data[1], data[2]])

                res = dbg.disassemble_list(0x0000, 0x0000)
                self.assertIsInstance(res, list)
                self.assertEqual(len(res), 1)

    def test_memdump(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dbg = self._debug(program=[data[0], data[1], data[2]])

                offset = 0x000F
                offset_width = ((offset | 0x000F) - ((offset & 0xFFF0)) + 1)

                t = self.memdump_str(
                    addr=(offset & 0xFFF0),
                    data=data,
                    suffix=True
                )

                v = t.format(
                    start=(offset & 0xFFF0),
                    stop=(offset | 0x000F),
                    offset="",
                    width=(offset_width * 3)
                )
                with unittest.mock.patch(
                    'sys.stdout', new_callable=io.StringIO
                ) as mock_stdout:
                    dbg.memdump(offset)
                    self.assertIn(mock_stdout.getvalue().strip(), v)

    def test_stackdump(self):
        for data in self.subtests:
            with self.subTest(data=data):
                dbg = self._debug(program=[data[0], data[1], data[2]])

                offset = 0x01FD
                offset_width = ((offset | 0x000F) - ((offset & 0xFFF0)) + 1)

                t = self.memdump_str(
                    addr=(offset & 0xFFF0),
                    suffix=True
                )

                v = t.format(
                    start=(offset & 0xFFF0),
                    stop=(offset | 0x000F),
                    offset="",
                    width=(offset_width * 3)
                )
                with unittest.mock.patch(
                    'sys.stdout', new_callable=io.StringIO
                ) as mock_stdout:
                    dbg.stackdump()
                    self.assertIn(mock_stdout.getvalue().strip(), v)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_suites
----------------------------------

Tests for `py65emu` module.
"""


import unittest
import unittest.mock

from py65emu.cpu import CPU
from py65emu.mmu import MMU
from py65emu.operation import OpCodes, Operation, UndefinedOperation


class TestOperation(unittest.TestCase):
    def setUp(self):
        pass

    def _cpu(self, program=None, pc=0x0000) -> CPU:
        self.c = CPU(
            mmu=MMU([(0x0, 0x10000, False, program, pc)]),
            pc=pc,
        )
        mmu = MMU(
            [
                (0x0000, 0x800),  # RAM
                (0x2000, 0x8),  # PPU
                (0x4000, 0x18),
                (0x8000, 0xC000, True, [], pc),  # ROM
            ]
        )
        self.c = CPU(mmu, pc)
        self.c.r.s = 0xFD  # Not sure why the stack starts here.
        return self.c

    def test_unknown_operation(self):
        self._cpu()
        opc = OpCodes(self.c)
        with self.assertRaises(UndefinedOperation):
            opc[0x125]

    def test_repr_on_operation(self):
        subtests = [
            # OPC   LO    HI
            (0xEA, 'EA: NOP '),
            (0x0A, '0A: ASL A '),
            (0xA9, 'A9: LDA #$BB '),
            (0xAD, 'AD: LDA $LLHH '),
            (0xAD, 'AD: LDA $LLHH '),
            (0xA5, 'A5: LDA $LL '),
            (0xB5, 'B5: LDA $LL, X '),
            (0xB6, 'B6: LDX $LL, Y '),
            (0xAD, 'AD: LDA $LLHH '),
            (0xBD, 'BD: LDA $LLHH, X '),
            (0xB9, 'B9: LDA $LLHH, Y '),
            (0xA1, 'A1: LDA ($LL, X) '),
            (0xB1, 'B1: LDA ($LL), Y '),
            (0x6C, '6C: JMP ($LLHH) '),
            (0x90, '90: BCC $BB [PC + $BB] '),
            (0xB0, 'B0: BCS $BB [PC + $BB] '),
        ]

        self._cpu()
        for data in subtests:
            with self.subTest(data=data):
                c = self.c.opcodes[data[0]]

                self.assertIsInstance(c, Operation)
                self.assertEqual(repr(c), data[1])

    def tearDown(self):
        pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_klaus_dormann
----------------------------------

Klaus Dormann's tests used for `py65emu` module.
"""


import os
import unittest
import copy

from py65emu.cpu import CPU, FlagBit
from py65emu.mmu import MMU
from py65emu.debug import Debug


class KlausDormann(unittest.TestCase):
    c: CPU

    def setUp(self) -> None:
        pass

    def _cpu(self, path: str, pc: int = 0x400) -> CPU:
        mmu = [
            (0, pc, False),
        ]
        self.c = CPU(MMU(mmu), pc)

        with open(path, "rb") as fp:
            self.c.mmu.addBlock(pc, 0x10000 - pc, False, fp)

        self.c.mmu.write(0xFFFC, (pc >> 8) & 0xFF)
        self.c.mmu.write(0xFFFD, pc & 0xFF)

        return self.c

    def tearDown(self) -> None:
        Debug.crash_dump(self.c)


class Decimal:
    cpu: CPU | None = None

    N1: int | None = None
    N2: int | None = None
    HA: int | None = None
    HNVZC: int | None = None
    DA: int | None = None
    DNVZC: int | None = None
    AR: int | None = None
    NF: int | None = None
    VF: int | None = None
    ZF: int | None = None
    CF: int | None = None
    ERROR: int | None = None
    N1L: int | None = None
    N1H: int | None = None
    N2L: int | None = None
    N2H: int | None = None

    def __init__(self, cpu: CPU | None = None):
        self.cpu = cpu

    def update(self) -> None:
        if self.cpu is None:
            return None

        # operands - register Y = carry in
        # 0x0000 1 byte
        self.N1 = self.cpu.mmu.read(0x0000)
        # 0x0001 1 byte
        self.N2 = self.cpu.mmu.read(0x0001)
        # binary result
        # 0x0002 1 byte
        self.HA = self.cpu.mmu.read(0x0002)
        # 0x0003 1 byte
        self.HNVZC = self.cpu.mmu.read(0x0003)
        # decimal result
        # 0x0004 1 byte
        self.DA = self.cpu.mmu.read(0x0004)
        # 0x0005 1 byte
        self.DNVZC = self.cpu.mmu.read(0x0005)
        # predicted results
        # 0x0006 1 byte
        self.AR = self.cpu.mmu.read(0x0006)
        # 0x0007 1 byte
        self.NF = self.cpu.mmu.read(0x0007)

        # 0x0008 1 byte
        self.VF = self.cpu.mmu.read(0x0008)
        # 0x0009 1 byte
        self.ZF = self.cpu.mmu.read(0x0009)
        # 0x000a 1 byte
        self.CF = self.cpu.mmu.read(0x000a)
        # 0x000b 1 byte
        self.ERROR = self.cpu.mmu.read(0x000b)
        # workspace
        # 0x000c 1 byte
        self.N1L = self.cpu.mmu.read(0x000c)
        # 0x000d 1 byte
        self.N1H = self.cpu.mmu.read(0x000d)
        # 0x000e 1 byte
        self.N2L = self.cpu.mmu.read(0x000e)
        # 0x000f 2 byte
        self.N2H = (
            self.cpu.mmu.read(0x000f) | (self.cpu.mmu.read(0x0010) << 8)
        )

    def __eq__(self, other):
        if not isinstance(other, Decimal):
            # don't attempt to compare against unrelated types
            return NotImplemented

        attr = [
            "N1", "N2",
            "HA", "HNVZC",
            "DA", "DNVZC",
            "AR", "NF", "VF", "ZF", "CF", "ERROR",
            "N1L", "N1H", "N2L", "N2H"
        ]
        for a in attr:
            self_value = getattr(self, a)
            other_value = getattr(other, a)

            if self_value is None:
                raise ValueError(f"L: '{a}' is None")

            if other_value is None:
                raise ValueError(f"R: '{a}' is None")

            if self_value != other_value:
                return False
        return True

    def __xor__(self, other):
        if not isinstance(other, Decimal):
            # don't attempt to compare against unrelated types
            return NotImplemented

        new = Decimal()

        attr = [
            "N1", "N2",
            "HA", "HNVZC",
            "DA", "DNVZC",
            "AR", "NF", "VF", "ZF", "CF", "ERROR",
            "N1L", "N1H", "N2L", "N2H"
        ]
        for a in attr:
            self_value = getattr(self, a)
            other_value = getattr(other, a)
            if self_value == other_value:
                continue
            result = self_value ^ other_value
            setattr(new, a, result)

        return new


@unittest.skip('Runs, but doesn\'t break')
class KlausDormannDecimal(KlausDormann):
    def setUp(self):
        super().setUp()
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            "6502_decimal_test.bin",
        )
        self.c = self._cpu(path, 0x200)

    def test(self):
        """
        Bruce Clark - Verify decimal mode behavior, modified by Klaus Dormann
        See https://github.com/Klaus2m5/6502_65C02_functional_tests
        """

        total_no_of_cycles = 0
        self.c.running = True

        comparison_value = Decimal(self.c)
        comparison_value.update()
        new_value = copy.copy(comparison_value)

        while self.c.running:
            self.c.step()

            new_value.update()
            if comparison_value != new_value:
                # res = comparison_value ^ new_value
                comparison_value = new_value

            # total_no_of_cycles += self.c.cc
            total_no_of_cycles += 1

            if self.c.running is not True:
                print("No. of cycles {}".format(total_no_of_cycles))
                self.assertEqual(
                    self.c.mmu.read(0x000b),
                    0,
                    "Test reported error"
                )
                break

            if self.c.r.pc == 0x0257:
                break

            self.assertLess(
                total_no_of_cycles,
                0x2500000,
                "Maximum number of loops exceeded"
            )

        print("No. of cycles {}".format(total_no_of_cycles))


@unittest.skip('Takes a long time to run!')
class KlausDormannFunctional(KlausDormann):
    def setUp(self):
        super().setUp()
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            "6502_functional_test.bin",
        )
        self.c = self._cpu(path)

    def test(self):
        """
        Klaus Dormann's Functional Test Program.
        See https://github.com/Klaus2m5/6502_65C02_functional_tests

        If a test is failing find the first test that fails. The tests are
        dumb, they do not catch error traps correctly.
        """

        data = {
            # Load Data
            0x0461: 0x01,
            # testing relative addressing with BEQ
            0x05AA: 0x02,
            # partial test BNE & CMP, CPX, CPY immediate
            0x05F1: 0x03,
            # testing stack operations PHA PHP PLA PLP
            0x0625: 0x04,
            # testing branch decisions BPL BMI BVC BVS BCC BCS BNE BEQ
            0x079F: 0x05,
            # test PHA does not alter flags or accumulator but PLA does
            0x089B: 0x06,
            # partial pretest EOR #
            0x08CF: 0x07,
            # PC modifying instructions except branches
            # (NOP, JMP, JSR, RTS, BRK, RTI)
            0x0919: 0x08,
            # Jump absolute
            0x096F: 0x09,
            # Jump indirect
            0x09AB: 0x0A,
            # Jump subroutine & return from subroutine
            0x09E2: 0x0B,
            # break & return from interrupt
            0x0A3A: 0x0C,
            # test set and clear flags CLC CLI CLD CLV SEC SEI SED
            0x0AE0: 0x0D,
            # testing index register increment/decrement and transfer
            # INX INY DEX DEY TAX TXA TAY TYA
            0x0DA6: 0x0E,
            # TSX sets NZ - TXS does not
            # This section also tests for proper stack wrap around.
            0x0E6F: 0x0F,
            # testing index register load & store LDY LDX STY STX all
            # addressing modes
            # LDX / STX - zp,y / abs,y
            0x0F2A: 0x10,
            # Indexed wraparound test (only zp should wrap)
            0x0F6C: 0x11,
            # LDY / STY - zp,x / abs,x
            0x1023: 0x12,
            # Indexed wraparound test (only zp should wrap)
            0x1063: 0x13,
            # LDX / STX - zp / abs / #
            0x1359: 0x14,
            # LDY / STY - zp / abs / #
            0x1653: 0x15,
            # Testing load / store accumulator LDA / STA all addressing modes
            # LDA / STA - zp,x / abs,x
            0x1704: 0x16,
            # LDA / STA - (zp),y / abs,y / (zp,x)
            0x181F: 0x17,
            # Indexed wraparound test (only zp should wrap)
            0x18C2: 0x18,
            # LDA / STA - zp / abs / #
            0x1B8C: 0x19,
            # testing bit test & compares BIT CPX CPY CMP all addressing modes
            # BIT - zp / abs
            0x1CE0: 0x1A,
            # CPX - zp / abs / #
            0x1DEE: 0x1B,
            # CPY - zp / abs / #
            0x1EFC: 0x1C,
            # CMP - zp / abs / #
            0x22E0: 0x1D,
            # testing shifts - ASL LSR ROL ROR all addressing modes
            # shifts - accumulator
            0x2424: 0x1E,
            # Shifts - zeropage
            0x25A4: 0x1F,
            # Shifts - absolute
            0x2748: 0x20,
            # Shifts - zp indexed
            0x28C8: 0x21,
            # Shifts - abs indexed
            0x2A6C: 0x22,
            # testing memory increment/decrement - INC DEC all addressing modes
            # zeropage
            0x2B16: 0x23,
            # absolute memory
            0x2BD0: 0x24,
            # zeropage indexed
            0x2C7E: 0x25,
            # memory indexed
            0x2D3C: 0x26,
            # Testing logical instructions - AND EOR ORA all addressing modes
            # AND
            0x2F34: 0x27,
            # EOR
            0x312C: 0x28,
            # OR
            0x3325: 0x29,
            # full binary add/subtract test
            # iterates through all combinations of operands and carry input
            # uses increments/decrements to predict result & result flags
            0x338A: 0x2A,
            # decimal add/subtract test
            # *** WARNING - tests documented behavior only! ***
            #   only valid BCD operands are tested, N V Z flags are ignored
            # iterates through all valid combinations of operands and carry
            # input uses increments/decrements to predict result & carry flag
            0x342E: 0x2B,
            # decimal/binary switch test
            # tests CLD, SED, PLP, RTI to properly switch between decimal &
            # binary opcode
            0x3486: 0xF0,
        }

        self.c.debug = True
        while 1:
            old_pc = self.c.r.pc
            self.c.step()

            if self.c.r.pc in data:
                self.assertEqual(self.c.r.a, data[self.c.r.pc])

            self.assertNotEqual(
                self.c.r.pc,
                old_pc,
                f'Catched Trap {old_pc:0<2x}'
            )

            self.assertLess(
                self.c.cc_total,
                0x262EE18,  # 40037912
                "Maximum number of loops exceeded"
            )


@unittest.expectedFailure
@unittest.skip('Not working without a IRQ/NMI generated')
class KlausDormannInterrupt(KlausDormann):
    def setUp(self):
        super().setUp()
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            "6502_interrupt_test.bin",
        )
        self.c = self._cpu(path)

    def test(self):
        """
        Klaus_Dormann's Interrupt Test Program.
        See https://github.com/Klaus2m5/6502_65C02_functional_tests

        This tests that the IRQ BRK and NMI all function correctly.
        """
        previous_interrupt_watch_value = 0
        total_no_of_cycles = 0

        """
        {
            # Load Data
            0x0424: True,
            # IRQ integrity test
            # test for clear flags seen in IRQ vector
            0x0501: True,
            # BRK integrity test
            # test for clear flags seen in IRQ vector
            0x05BF: True,
            # NMI integrity test
            # test for clear flags seen in NMI vector
            0x069D: True,
            # test IRQ & NMI with interrupts disabled
            0x06CF: True,
            # test overlapping NMI, IRQ & BRK
            0x0700: True,
            # 65C02 - WAI with interrupts disabled
            0x071a: False,
            # 65C02 - WAI with interrupts enabled
            0x0737: False,
            # 65C02 - manual test for the STP opcode of the 65c02
            0x073a: False,
        }
        """

        while 1:
            interrupt_watch = self.c.mmu.read(0xBFFC)
            # This is used to simulate the edge triggering of an NMI.
            # If we didn't do this we would get stuck in a loop forever
            if interrupt_watch != previous_interrupt_watch_value:
                previous_interrupt_watch_value = interrupt_watch
                if (interrupt_watch & 2) != 0:
                    self.c.trigger_nmi = True

            if (
                (
                    self.c.r.getFlag(FlagBit.I) is False
                ) and (
                    (interrupt_watch & 1) != 0
                )
            ):
                self.c.interruptRequest()

            old_pc = self.c.r.pc
            self.c.step()

            if self.c.r.pc == 0x0700:
                # Success
                break

            self.assertNotEqual(
                self.c.r.pc,
                old_pc,
                f'Catched Trap {old_pc:0<2x}'
            )

            self.assertLess(
                total_no_of_cycles,
                0x186A0,
                "Maximum number of loops exceeded"
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_mmu
----------------------------------
"""

import os
import unittest

from py65emu.mmu import MMU, MemoryRangeError, ReadOnlyError


class TestMMU(unittest.TestCase):
    def setUp(self):
        pass

    def test_create_empty(self):
        MMU([])

    def test_create(self):
        MMU([(0, 128, False, None)])

        MMU([(0, 128, False, None), (128, 128, True, None)])

    def test_create_with_list(self):
        m = MMU([(0, 128, False, [1, 2, 3])])

        self.assertEqual(m.blocks[0].get(0), 1)
        self.assertEqual(m.blocks[0].get(2), 3)
        self.assertEqual(m.blocks[0].get(3), 0)

    def test_create_with_file(self):
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "files",
            "test_load_file.bin"
        )

        with open(path, "rb") as f:
            m = MMU([(0, 128, True, f)])

        self.assertEqual(m.blocks[0].get(0), 0xA9)

    def test_create_overlapping(self):
        with self.assertRaises(MemoryRangeError):
            MMU([(0, 129), (128, 128)])

    def test_addBlock(self):
        m = MMU([])
        m.addBlock(0, 128, False, None)
        m.addBlock(128, 128, True, None)

    def test_addBlock_overlapping(self):
        m = MMU([])
        m.addBlock(128, 128)
        with self.assertRaises(MemoryRangeError):
            m.addBlock(0, 129)
        with self.assertRaises(MemoryRangeError):
            m.addBlock(255, 128)

    def test_write(self):
        m = MMU([(0, 128)])
        m.write(16, 25)
        self.assertEqual(m.blocks[0].get(16), 25)

    def test_write_multiple_blocks(self):
        m = MMU([(0, 128), (1024, 128)])
        m.write(16, 25)
        self.assertEqual(m.blocks[0][16], 25)
        m.write(1056, 55)
        self.assertEqual(m.blocks[1][32], 55, m.blocks[1])

    def test_write_readonly(self):
        m = MMU([(0, 16, True), (16, 16), (32, 16, True)])
        with self.assertRaises(ReadOnlyError):
            m.write(8, 1)
        m.write(20, 1)
        with self.assertRaises(ReadOnlyError):
            m.write(40, 1)

    def test_read(self):
        m = MMU([(0, 128)])
        m.write(16, 5)
        m.write(64, 111)
        self.assertEqual(5, m.read(16))
        self.assertEqual(111, m.read(64))

    def test_index_error(self):
        m = MMU([(0, 128)])
        with self.assertRaises(IndexError):
            m.write(-1, 0)
        with self.assertRaises(IndexError):
            m.write(128, 0)
        with self.assertRaises(IndexError):
            m.read(-1)
        with self.assertRaises(IndexError):
            m.read(128)
        with self.assertRaises(IndexError):
            b = m.getBlock(127)
            _ = b[256]
        with self.assertRaises(IndexError):
            b = m.getBlock(127)
            b[256] = 0
        with self.assertRaises(IndexError):
            b = m.getBlock(127)
            b.get(256)
        with self.assertRaises(IndexError):
            b = m.getBlock(127)
            b.set(256, 0)

    def test_that_repr_is_strin(self):
        m = MMU([(0, 128)])
        b = m.getBlock(120)
        s = repr(b)
        self.assertIsInstance(s, str)

    def test_reset(self):
        m = MMU([(0, 16, True), (16, 16, False)])
        m.blocks[0][0] = 5
        m.write(16, 10)
        m.reset()
        self.assertEqual(m.read(0), 5)
        self.assertEqual(m.read(16), 0)

    def tearDown(self):
        pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_mmu
----------------------------------
"""

import unittest

from py65emu.cpu import FlagBit, Registers


class TestRegisters(unittest.TestCase):
    def setUp(self):
        pass

    def test_flags_enum(self):
        r = Registers()
        r.setFlag(FlagBit.N)
        self.assertEqual(r.p, 0b10100100)
        r.setFlag(FlagBit.Z)
        self.assertEqual(r.p, 0b10100110)
        r.clearFlag(FlagBit.N)
        self.assertEqual(r.p, 0b00100110)
        self.assertTrue(r.getFlag(FlagBit.Z))
        r.setFlag(FlagBit.Z, False)
        self.assertFalse(r.getFlag(FlagBit.Z))
        r.setFlag(FlagBit.Z, False)
        self.assertFalse(r.getFlag(FlagBit.Z))

    def test_flags_string(self):
        r = Registers()
        r.setFlag("N")
        self.assertEqual(r.p, 0b10100100)
        r.setFlag("Z")
        self.assertEqual(r.p, 0b10100110)
        r.clearFlag("N")
        self.assertEqual(r.p, 0b00100110)
        self.assertTrue(r.getFlag("Z"))
        r.setFlag("Z", False)
        self.assertFalse(r.getFlag("Z"))
        r.setFlag("Z", False)
        self.assertFalse(r.getFlag("Z"))

    def tearDown(self):
        pass


if __name__ == "__main__":
    unittest.main()
